/*** * **** * ***** ********* ** ****** ***** *** ***** ********
 **
 ** Copyright (c) 2005-2010, SkeTo Project
 ** All rights reserved.
 **
 ** This file is part of the SkeTo Library. This library is free
 ** software; you can redistribute it and/or modify it under the
 ** terms of the MIT License (see COPYING in the package).
 ** 
 ** This library is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 **
 *** * **** * ***** ********* ** ****** ***** *** ***** ********/

#ifndef __LIST_SKELETONS_COMMON_TPP__
#define __LIST_SKELETONS_COMMON_TPP__
#include "transporter.h"
namespace sketo {
  /*
   * This function performs "global reduce" on the values specified as pval.
   * @param oplus An associative operator used in reduce. This opertor should be
   *              a function object of type "A oplus(A, A)".
   * @param pval The address of the value reduced on each process.
   *             The result of the global reduce will be stored on this.
   */
  template <typename A, typename OP>
  void
  _impl::ls_adpt::global_reduce(const OP& oplus,
                                A* pval)
  {
    for (int stage = 1; stage < sketo::procs; stage <<= 1) {
      int target = sketo::rank ^ stage;
      if (target < sketo::rank) {
        _impl::transporter::send<A>(pval, 1, target, TAG_REDUCE);
        break;
      }
      else if (target < sketo::procs) {
        A received_data;
        _impl::transporter::recv<A>(&received_data, 1, target, TAG_REDUCE);
        *pval = oplus(*pval, received_data);
      }
      else {
        // do nothing
      }
    }
    _impl::transporter::bcast<A>(pval, 1, 0);
  }

  /*
   * This function performs "global scan" on the values specified as pfirst
   * and ptotal. The first value (pfirst) is assumed to be a dummy value.
   * @param oplus An associative operator used in scan. This operator should be
   *              a function object of type "A oplus(A, A)".
   * @param pfirst The address of the value on which the result of scan is stored.
   * @param ptotal The address of the input value that is result of local reduce.
   */
  template <typename A, typename OP>
  void
  _impl::ls_adpt::global_scan(const OP& oplus,
                              A* pfirst,
                              A* ptotal)
  {
    bool is_firstval_set = false;
    for (int stage = 1; stage < sketo::procs; stage <<= 1) {
      const int target = sketo::rank ^ stage;
      if (target >= sketo::procs) continue;
      A received_val;
      _impl::transporter::sendrecv<A>(ptotal, &received_val, 1, target, TAG_SCAN1);
      if (target > sketo::rank) {
        *ptotal = oplus(*ptotal, received_val);
      } else {
        *ptotal = oplus(received_val, *ptotal);
        if (is_firstval_set) {
          *pfirst = oplus(received_val, *pfirst);
        } else {
          *pfirst = received_val; is_firstval_set = true;
        }
      }
    }
  }

  /*
   * This function performs "global scanr" on the values specified as plast
   * and ptotal. The last value (plast) is assumed to be a dummy value.
   * @param oplus An associative operator used in scan. This operator should be
   *              a function object of type "A oplus(A, A)".
   * @param plast The address of the value on which the result of scan is stored.
   * @param ptotal The address of the input value that is result of local reduce.
   */
  template <typename A, typename OP>
  void
  _impl::ls_adpt::global_scanr(const OP& oplus,
                               A* plast,
                               A* ptotal)
  {
    bool is_lastval_set = false;
    const int vrank = sketo::procs - sketo::rank - 1;
    
    for (int stage = 1; stage < sketo::procs; stage <<= 1) {
      const int vtarget = vrank ^ stage;
      const int target = sketo::procs - vtarget - 1;
      if (target < 0) continue;
      A received_val;
      _impl::transporter::sendrecv<A>(ptotal, &received_val, 1, target, TAG_SCAN1);
      if (vtarget > vrank) {
        *ptotal = oplus(received_val, *ptotal);
      } else {
        *ptotal = oplus(*ptotal, received_val);
        if (is_lastval_set) {
          *plast = oplus(*plast, received_val);
        } else {
          *plast = received_val; is_lastval_set = true;
        }
      }
    }
  }
  
  template <typename OPLUS, typename A>
  dist_list<A>
  _impl::ls_adpt::postscan(const OPLUS &oplus,
                           const dist_list<A> &as)
  {
    // local reduce
    A total_sum;
    {
      const int n = as.get_local_size();
      total_sum = as.local_get(0);
      for (int i = 1; i < n; ++i) {
        total_sum = oplus(total_sum, as.local_get(i));
      }
    }

    // global_scan
    A first_val;
    global_scan(oplus, &first_val, &total_sum);

    dist_list<A> bs(as.get_global_size());
    // local_scan
    {
      const int n = as.get_local_size();
      A* const bs_buf = bs.data->buf;
      bs_buf[0] = (rank == 0) ? as.local_get(0) : oplus(first_val, as.local_get(0));
      for (int i = 1; i < n; ++i) {
        bs_buf[i] = oplus(bs_buf[i-1], as.local_get(i));
      }
    }

    return bs;
  }

  template <typename OPLUS, typename A>
  dist_list<A>
  _impl::ls_adpt::postscanr(const OPLUS &oplus,
                            const dist_list<A> &as)
  {
    // local reduce
    A total_sum;
    {
      const int n = as.get_local_size();
      total_sum = as.local_get(0);
      for (int i = 1; i < n; ++i) {
        total_sum = oplus(total_sum, as.local_get(i));
      }
    }

    // global_scanr
    A last_val;
    global_scanr(oplus, &last_val, &total_sum);

    dist_list<A> bs(as.get_global_size());
    // local_scan
    {
      A* const bs_buf = bs.data->buf;
      const int n = as.get_local_size();
      bs_buf[n-1] = (rank == procs - 1) ? as.local_get(n-1) : oplus(as.local_get(n-1), last_val);
      for (int i = n - 2; i >= 0; --i) {
        bs_buf[i] = oplus(as.local_get(i), bs_buf[i+1]);
      }
    }

    return bs;
  }

  template <typename F, typename G, typename OMINUS, typename OPLUS,
            typename A, typename B, typename C>
  dist_list<B>
  _impl::ls_adpt::gscanl(const F &f,
                         const B &c,
                         const G &g,
                         const OMINUS &ominus,
                         const OPLUS &oplus,
                         const dist_list<A> &as)
  {
    // local map and reduce
    C total_sum;
    {
      const int n = as.get_local_size();
      total_sum = g(as.local_get(0));
      for (int i = 1; i < n; ++i) {
        total_sum = oplus(total_sum, g(as.local_get(i)));
      }
    }

    // global_scan
    C first_val;
    global_scan(oplus, &first_val, &total_sum);

    dist_list<B> bs(as.get_global_size());
    // local scan
    {
      B* const bs_buf = bs.data->buf;
      const int n = as.get_local_size();
      bs_buf[0] = (rank == 0) ? c : ominus(c, first_val);
      for (int i = 1; i < n; ++i) {
        bs_buf[i] = f(bs_buf[i-1], as.local_get(i-1));
      }
    }

    return bs;
  }

  template <typename F, typename G, typename OMINUS, typename OPLUS,
            typename A, typename B, typename C>
  dist_list<B>
  _impl::ls_adpt::gscanr(const F &f,
                         const B &c,
                         const G &g,
                         const OMINUS &ominus,
                         const OPLUS &oplus,
                         const dist_list<A>& as)
  {
    // local map and reduce
    C total_sum;
    {
      const int n = as.get_local_size();
      total_sum = g(as.local_get(n-1));
      for (int i = n-2; i >= 0; --i) {
        total_sum = oplus(total_sum, g(as.local_get(i)));
      }
    }

    // global_scan
    C last_val;
    global_scanr(sketo::functions::swap(oplus), &last_val, &total_sum);

    dist_list<B> bs(as.get_global_size());
    // local scan
    {
      B* const bs_buf = bs.data->buf;
      const int n = as.get_local_size();
      bs_buf[n-1] = (rank == sketo::procs - 1) ? c : ominus(c, last_val);
      for (int i = n-2; i >= 0; --i) {
        bs_buf[i] = f(bs_buf[i+1], as.local_get(i+1));
      }
    }

    return bs;
  }

}


#endif // __LIST_SKELETONS_COMMON_TPP__

