/*** * **** * ***** ********* ** ****** ***** *** ***** ********
 **
 ** Copyright (c) 2005-2010, SkeTo Project
 ** All rights reserved.
 **
 ** This file is part of the SkeTo Library. This library is free
 ** software; you can redistribute it and/or modify it under the
 ** terms of the MIT License (see COPYING in the package).
 ** 
 ** This library is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 **
 *** * **** * ***** ********* ** ****** ***** *** ***** ********/

/**
 * @file list_skeletons_with_fusion.h
 * @brief An implmentation of list skeletons with fusion optimization.
 */

#ifndef __LIST_SKELETONS_WITH_FUSION_H__
#define __LIST_SKELETONS_WITH_FUSION_H__
#include <cassert>
#include "transporter.h"
namespace sketo {
  namespace _impl {
    // Classes for expression templates
    template <typename F>
    struct ls_genobj {
      F f;
      int global_size;
      int local_begin;
      int local_size;
      typedef typename result_of<F(int)>::type element_type;
      ls_genobj(const F &f, int global_size) : f(f), global_size(global_size) {
        sketo::_impl::dist_list_buffer<void>::compute_local_array(global_size, sketo::procs, sketo::rank,
                                                                  &local_begin, &local_size);
      }
      int get_global_size() const { return global_size; }
      int get_local_size() const { return local_size; }
      element_type local_get(int i) const { return f(local_begin + i); }
      template <typename T>
      void check_overwrite(_impl::dist_list_buffer<T>* target,
                           int *refcount, bool *no_deps, bool *child_no_deps, bool *scan_shift) const;
      dist_list<element_type>* ow_impl(dist_list<element_type>*) const { assert(false); return NULL; }
    };

    template <typename F, typename AS>
    struct ls_mapobj {
      F f;
      const AS as;
      typedef typename result_of<F(typename AS::element_type)>::type element_type;
      ls_mapobj(const F &f, const AS &as) : f(f), as(as) { }
      int get_global_size() const { return as.get_global_size(); }
      int get_local_size() const { return as.get_local_size(); }
      element_type local_get(int i) const { return f(as.local_get(i)); }
      template <typename T>
      void check_overwrite(_impl::dist_list_buffer<T>* target,
                           int *refcount, bool *no_deps, bool *child_no_deps, bool *scan_shift) const;
      dist_list<element_type>* ow_impl(dist_list<element_type>*) const { assert(false); return NULL; }
    };

    template <typename F, typename AS, typename BS>
    struct ls_zipwobj {
      F f;
      const AS as;
      const BS bs;
      typedef typename result_of<F(typename AS::element_type, typename BS::element_type)>::type element_type;
      ls_zipwobj(const F &f, const AS &as, const BS &bs) : f(f), as(as), bs(bs) { }
      int get_global_size() const { return as.get_global_size(); }
      int get_local_size() const { return as.get_local_size(); }
      element_type local_get(int i) const { return f(as.local_get(i), bs.local_get(i)); }
      template <typename T>
      void check_overwrite(_impl::dist_list_buffer<T>* target,
                           int *refcount, bool *no_deps, bool *child_no_deps, bool *scan_shift) const;
      dist_list<element_type>* ow_impl(dist_list<element_type>*) const { assert(false); return NULL; }
    };

    template <typename AS, typename A>
    struct ls_shiftlobj {
      const AS as;
      const A fromR;
      typedef A element_type;
      ls_shiftlobj(const AS &as, const A &fromR) : as(as), fromR(fromR) { }
      int get_global_size() const { return as.get_global_size(); }
      int get_local_size() const { return as.get_local_size(); }
      element_type local_get(int i) const {
        if (i == get_local_size() - 1) return fromR;
        else return as.local_get(i + 1);
      }
      template <typename T>
      void check_overwrite(_impl::dist_list_buffer<T>* target,
                           int *refcount, bool *no_deps, bool *child_no_deps, bool *scan_shift) const;
      dist_list<element_type>* ow_impl(dist_list<element_type>*) const;
    };

    template <typename AS, typename A>
    class ls_shiftrobj {
      const AS as;
      const A fromL;
    public:
      typedef A element_type;
      ls_shiftrobj(const AS &as, const A &fromL) : as(as), fromL(fromL) { }
      int get_global_size() const { return as.get_global_size(); }
      int get_local_size() const { return as.get_local_size(); }
      element_type local_get(int i) const {
        if (i == 0) return fromL;
        else return as.local_get(i - 1);
      }
      template <typename T>
      void check_overwrite(_impl::dist_list_buffer<T>* target,
                           int *refcount, bool *no_deps, bool *child_no_deps, bool *scan_shift) const;
      dist_list<element_type>* ow_impl(dist_list<element_type>*) const;
    };

    template <typename OP, typename AS, typename A>
    struct ls_scanobj {
      const OP op;
      const AS as;
      const A e;
      A* const last;
      mutable ls_mapobj<std::binder1st<OP>, dist_list<A> > *subresult;
      typedef A element_type;
      ls_scanobj(const OP &op, const AS &as, const A &e, A* last) : op(op), as(as), e(e), last(last), subresult(NULL) { }
      ~ls_scanobj() { if (subresult) delete subresult; }
      int get_global_size() const { return as.get_global_size(); }
      int get_local_size() const { return as.get_local_size(); }
      element_type local_get(int i) const {
        if (!subresult) subeval();
        return subresult->local_get(i);
      }
      template <typename T>
      void check_overwrite(_impl::dist_list_buffer<T>* target,
                           int *refcount, bool *no_deps, bool *child_no_deps, bool *scan_shift) const;
      dist_list<element_type>* ow_impl(dist_list<element_type>* bs) const;
    private:
      void subeval() const {
        dist_list<A> subarray(as.get_global_size());
        A first_val;
        _impl::ls_adpt::scanobj_subeval(&subarray, &first_val, op, as, e, last);
        subresult = new ls_mapobj<std::binder1st<OP>, dist_list<A> >(std::bind1st(op, first_val), subarray);
      }
    };

    template <typename OP, typename AS, typename A>
    struct ls_scanrobj {
      const OP op;
      const AS as;
      const A e;
      A* first;
      mutable ls_mapobj<std::binder2nd<OP>, dist_list<A> > *subresult;
      typedef A element_type;
      ls_scanrobj(const OP &op, const AS &as, const A &e, A* first) : op(op), as(as), e(e), first(first), subresult(NULL) { }
      ~ls_scanrobj() { if (subresult) delete subresult; }
      int get_global_size() const { return as.get_global_size(); }
      int get_local_size() const { return as.get_local_size(); }
      element_type local_get(int i) const {
        if (!subresult) subeval();
        return subresult->local_get(i);
      }
      template <typename T>
      void check_overwrite(_impl::dist_list_buffer<T>* target,
                           int *refcount, bool *no_deps, bool *child_no_deps, bool *scan_shift) const;
      dist_list<element_type>* ow_impl(dist_list<element_type>*) const;
    private:
      void subeval() const {
        dist_list<A> subarray(as.get_global_size());
        A last_val;
        _impl::ls_adpt::scanrobj_subeval(&subarray, &last_val, op, as, e, first);
        subresult = new ls_mapobj<std::binder2nd<OP>, dist_list<A> >(std::bind2nd(op, last_val), subarray);
      }
    };
      
//     template <typename OP, typename AS>
//     class ls_postscanobj {
//       const OP op;
//       const AS as;
//     public:
//       typedef typename AS::element_type element_type;
//       ls_postscanobj(const OP &op, const AS &as) : op(op), as(as) { }
//       int get_global_size() const { return as.get_global_size(); }
//     };

//     template <typename OP, typename AS>
//     class ls_postscanrobj {
//       const OP op;
//       const AS as;
//     public:
//       typedef typename AS::element_type element_type;
//       ls_postscanrobj(const OP &op, const AS &as) : op(op), as(as) { }
//       int get_global_size() const { return as.get_global_size(); }
//     };

  }
  
  namespace list_skeletons {
    /**
     * generate :: int -> (int -> a) -> [a]
     */
    template <typename F>
    _impl::ls_genobj<F>
    generate(int global_size,
             const F& f) {
      return _impl::ls_genobj<F>(f, global_size);
    }

    /**
     * map :: (a -> b) -> [a] -> [b]
     */
    template <typename F, typename AS>
    _impl::ls_mapobj<F, AS>
    map(F f,
        const AS& as) {
      return _impl::ls_mapobj<F, AS>(f, as);
    }

    /**
     * map_with_index :: (int -> a -> b) -> [a] -> [b]
     */
    template <typename F, typename AS>
    _impl::ls_zipwobj<F, _impl::ls_genobj<sketo::functions::identity<int> >, AS >
    map_with_index(const F& f,
                   const AS& as) {
      return _impl::ls_zipwobj<F, _impl::ls_genobj<sketo::functions::identity<int> >, AS >
        (f, _impl::ls_genobj<sketo::functions::identity<int> >(sketo::functions::identity<int>(), as.get_global_size()), as);
    }

    /**
     * zip :: [a] -> [b] -> [(a, b)]
     */
    template <typename AS, typename BS>
    _impl::ls_zipwobj<sketo::functions::mkpair<typename AS::element_type, typename BS::element_type>, AS, BS>
    zip(const AS& as,
        const BS& bs) {
      return _impl::ls_zipwobj<sketo::functions::mkpair<typename AS::element_type, typename BS::element_type>, AS, BS>
        (sketo::functions::mkpair<typename AS::element_type, typename BS::element_type>(), as, bs);
    }

    /**
     * zipwith :: (a -> b -> c) -> [a] -> [b] -> [c]
     */
    template <typename F, typename AS, typename BS>
    _impl::ls_zipwobj<F, AS, BS>
    zipwith(const F& f,
            const AS& as,
            const BS& bs) {
      return _impl::ls_zipwobj<F, AS, BS>(f, as, bs);
    }

    /**
     * reduce :: (a -> a -> a) -> [a] -> [a]
     */
    template <typename OPLUS, typename AS>
    typename AS::element_type
    reduce(const OPLUS &oplus,
           const AS &as) {
      return _impl::ls_adpt::reduce<OPLUS, AS, typename AS::element_type>(oplus, as);
    }

    /**
     * scan :: (a -> a -> a) -> (a, [a]) -> ([a], a)
     */
    template <typename OPLUS, typename AS>
    _impl::ls_scanobj<OPLUS, AS, typename AS::element_type>
    scan(const OPLUS &oplus,
         typename OPLUS::first_argument_type e,
         const AS& as,
         typename AS::element_type *last_val = NULL) {
      return _impl::ls_scanobj<OPLUS, AS, typename AS::element_type>(oplus, as, e, last_val);
    }

    /**
     * scanr :: (a -> a -> a) -> ([a], a) -> (a, [a])
     */
    template <typename OPLUS, typename AS>
    _impl::ls_scanrobj<OPLUS, AS, typename AS::element_type>
    scanr(const OPLUS &oplus,
          typename OPLUS::second_argument_type e,
          const AS& as,
          typename AS::element_type *first_val = NULL) {
      return _impl::ls_scanrobj<OPLUS, AS, typename AS::element_type>(oplus, as, e, first_val);
    }

//     /**
//      * postscan :: (a -> a -> a) -> [a] -> [a]
//      */
//     template <typename OPLUS, typename AS>
//     _impl::ls_postscanobj<OPLUS, AS>
//     postscan(const OPLUS &oplus,
// 	     const AS& as) {
//       return _impl::ls_postscanobj<OPLUS, AS>(oplus, as);
//     }

//     /**
//      * postscanr :: (a -> a -> a) -> [a] -> [a]
//      */
//     template <typename OPLUS, typename AS>
//     _impl::ls_postscanrobj<OPLUS, AS>
//     postscanr(const OPLUS &oplus,
// 	      const AS& as) {
//       return _impl::ls_postscanrobj<OPLUS, AS>(oplus, as);
//     }
    template <typename OPLUS, typename AS>
    dist_list<typename AS::element_type>
    postscan(const OPLUS &oplus,
             const AS& as) {
      dist_list<typename AS::element_type> as2 = as;
      return _impl::ls_adpt::postscan(oplus, as2);
    }

    template <typename OPLUS, typename AS>
    dist_list<typename AS::element_type>
    postscanr(const OPLUS &oplus,
              const AS& as) {
      dist_list<typename AS::element_type> as2 = as;
      return _impl::ls_adpt::postscanr(oplus, as2);
    }
    
    /**
     * shiftl :: ([a], a) -> (a, [a])
     */
    template <typename AS, typename A>
    _impl::ls_shiftlobj<AS, A>
    shiftl(const A &rightmost, 
           const AS &as,
           A *leftmost = NULL) {
      A fromR;
      _impl::ls_adpt::shiftl_edge(as.local_get(0), &fromR);

      if (sketo::rank == sketo::procs - 1) {
        fromR = rightmost;
      }

      // share leftmost if specified
      if (leftmost) {
        *leftmost = as.local_get(0);
        _impl::transporter::bcast<A>(leftmost, 1, 0);
      }

      return _impl::ls_shiftlobj<AS, A>(as, fromR);
    }
    
    /**
     * shiftr :: (a, [a]) -> ([a], a)
     */
    template <typename AS, typename A>
    _impl::ls_shiftrobj<AS, A>
    shiftr(const A &leftmost,
           const AS &as,
           A *rightmost = NULL) {
      A fromL;
      _impl::ls_adpt::shiftr_edge(as.local_get(as.get_local_size()-1), &fromL);

      if (sketo::rank == 0) {
        fromL = leftmost;
      }

      // share rightmost if specified
      if (rightmost) {
        *rightmost = as.local_get(as.get_local_size()-1);
        _impl::transporter::bcast<A>(rightmost, 1, 0);
      }

      return _impl::ls_shiftrobj<AS, A>(as, fromL);
    }

    /**
     * gscanl :: (b -> a -> b) -> b -> (a -> c) -> (c -> c -> c) -> (b -> c -> b) -> [a] -> [b]
     */
    template <typename F, typename G, typename OMINUS, typename OPLUS,
              typename B, typename AS>
    dist_list<B>
    gscanl(const F &f,
           const B &c,
           const G &g,
           const OMINUS &ominus,
           const OPLUS &oplus,
           const AS& as)
    {
      dist_list<typename AS::element_type> as2 = as;
      return _impl::ls_adpt::gscanl<F, G, OMINUS, OPLUS, typename AS::element_type, B, typename result_of<G(typename AS::element_type)>::type>(f, c, g, ominus, oplus, as2);
    }

    /**
     * gscanr :: (b -> a -> b) -> b -> (a -> c) -> (c -> c -> c) -> (b -> c -> b) -> [a] -> [b]
     */
    template <typename F, typename G, typename OMINUS, typename OPLUS,
              typename B, typename AS>
    dist_list<B>
    gscanr(const F &f,
           const B &c,
           const G &g,
           const OMINUS &ominus,
           const OPLUS &oplus,
           const AS &as)
    {
      dist_list<typename AS::element_type> as2 = as;
      return _impl::ls_adpt::gscanr<F, G, OMINUS, OPLUS, typename AS::element_type, B, typename result_of<G(typename AS::element_type)>::type>(f, c, g, ominus, oplus, as2);
    }

  }
}

#include "list_skeletons_with_fusion.tpp"

#endif // __LIST_SKELETONS_WITH_FUSION_H__
