/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2010, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/

#ifndef __MATRIX_SKELETONS_COMMON_TPP__
#define __MATRIX_SKELETONS_COMMON_TPP__

#include "transporter.h"

namespace sketo {
 
  /*
   * This function performs "global reduce" on the values specified as pval.
   * Note that the reduction is carried out with interleaving horizontal and
   * vertical divisions.
   * @param oplus  An associative operator used in reduce. This opertor should 
   *               be a function object of type "A oplus(A, A)".
   *               This shuold have the abide property with otimes.
   * @param otimes An associative operator used in reduce. This opertor should
   *               be a function object of type "A otimes(A, A)".
   *               This shuold have the abide property with oplus.
   * @param pval   The address of the value reduced on each process.
   *               The result of the global reduce will be stored on this.
   */
  template <typename A, typename OP, typename OT>
  void
  _impl::ms_adpt::global_reduce(const OP& oplus,
                                const OT& otimes,
                                A* pval,
                                const msize pa)
  {
    const int iprocs = pa.i;
    const int jprocs = pa.j;

    mindex rs = _impl::dist_matrix_buffer<A>::compute_ranks(pa, sketo::rank);
    const int irank = rs.i;
    const int jrank = rs.j;

    const int ij = iprocs > jprocs ? iprocs : jprocs;
    
    for (int stage = 1; stage < ij; stage <<=1) {
      int itarget = irank ^ stage;
      
      if (itarget < irank) {
        int target = _impl::dist_matrix_buffer<A>::compute_rank(pa, mindex(itarget, jrank));
        _impl::transporter::send<A>(pval, 1, target, TAG_REDUCE);
        break;
      } else if(itarget < iprocs){
        A received_data;
        int target = _impl::dist_matrix_buffer<A>::compute_rank(pa, mindex(itarget, jrank));
        _impl::transporter::recv<A>(&received_data, 1, target, TAG_REDUCE);
        *pval = oplus(*pval, received_data);
      }
      int jtarget = jrank ^ stage;
      
      if (jtarget < jrank) {
        int target = _impl::dist_matrix_buffer<A>::compute_rank(pa, mindex(irank, jtarget));        
        _impl::transporter::send<A>(pval, 1, target, TAG_REDUCE);
        break;
      } else if(jtarget < jprocs){
        A received_data;
        int target = _impl::dist_matrix_buffer<A>::compute_rank(pa, mindex(irank, jtarget));        
        _impl::transporter::recv<A>(&received_data, 1, target, TAG_REDUCE);
        
        *pval = otimes(*pval, received_data);
      }
    }
    _impl::transporter::bcast<A>(pval, 1, 0);
  }



  /* implementation of the simple loop reduction */
  template <typename OPLUS, typename OTIMES, typename AS, typename A>
  A
  _impl::ms_impl_select::ReduceSimple::reduce
  (const OPLUS &oplus, const OTIMES &otimes, const AS &as)
  {
    // local reduce
    A val = as.local_get(mindex(0, 0));
    const int nj = as.get_local_size().j;
    for (int j = 1; j < nj; ++j) {
      val = otimes(val, as.local_get(mindex(0, j)));
    }
      
    const int ni = as.get_local_size().i;
    for (int i = 1; i < ni; ++i) {
      A valj = as.local_get(mindex(i, 0));
      const int nj = as.get_local_size().j;
      for (int j = 1; j < nj; ++j) {
        valj = otimes(valj, as.local_get(mindex(i, j)));
      }
      val = oplus(val, valj);
    }
    // global reduction
    msize pa = sketo::_impl::dist_matrix_buffer<A>::compute_process_arrangement(as.get_global_size(), sketo::procs);
    _impl::ms_adpt::global_reduce(oplus, otimes, &val, pa);
    return val;
  }

  /* implementation of the divide-and-conquer style reduction */
  template <typename OPLUS, typename OTIMES, typename AS, typename A>
  A
  _impl::ms_impl_select::ReduceDC::reduce
  (const OPLUS &oplus, const OTIMES &otimes, const AS &as)
  {
            
    // local reduce
    const int m = as.get_local_size().i;
    const int n = as.get_local_size().j;
      
    const int nm = n > m ? n : m;
    A* idat = new A[m * n];
    {  // the first phase reads data from the given matrix 'as'.
      int stage = 1;
      const int stage2 = stage << 1;
      for (int i = 0; i < m; i+=stage2) {
        for (int j = 0; j < n; j+=stage2) {
          const int i2 = i ^ stage;
          const int j2 = j ^ stage;
          if (i2 < m) {
            if(j2 < n) {
              idat[i * n + j] = otimes(oplus(as.local_get(mindex(i, j)), as.local_get(mindex(i2, j))), oplus(as.local_get(mindex(i, j2)), as.local_get(mindex(i2, j2))));
            } else {
              idat[i * n + j] = oplus(as.local_get(mindex(i, j)), as.local_get(mindex(i2, j)));
            }
          } else {
            if(j2 < n) {
              idat[i * n + j] = otimes(as.local_get(mindex(i, j)), as.local_get(mindex(i, j2)));
            } else {
              idat[i * n + j] = as.local_get(mindex(i, j));
            }
          }
        }
      }
    }
    // the second phase reads data from the buffer 'idat'.
    for (int stage = 2; stage < nm; stage <<=1) {
      const int stage2 = stage << 1;
      for (int i = 0; i < m; i+=stage2) {
        for (int j = 0; j < n; j+=stage2) {
          const int i2 = i ^ stage;
          const int j2 = j ^ stage;
          if (i2 < m) {
            if(j2 < n) {
              idat[i * n + j] = otimes(oplus(idat[i * n + j], idat[i2 * n + j]), oplus(idat[i * n + j2], idat[i2 * n + j2]));
            } else {
              idat[i * n + j] = oplus(idat[i * n + j], idat[i2 * n + j]);
            }
          } else {
            if(j2 < n) {
              idat[i * n + j] = otimes(idat[i * n + j], idat[i * n + j2]);
            } else {
              // idat[i * n + j] = idat[i * n + j];
            }
          }
        }
      }
    }
    A val = idat[0];
    delete [] idat;
    // global reduction
    msize pa = sketo::_impl::dist_matrix_buffer<A>::compute_process_arrangement(as.get_global_size(), sketo::procs);
    _impl::ms_adpt::global_reduce(oplus, otimes, &val, pa);
    return val;
  }
 

  /*
   * This function performs multiple one-direction global scans at once.
   * 
   * @param oplus    An associative operator used in scan. This opertor should 
   *                 be a function object of type "A oplus(A, A)".
   * @param n        The nubmer of scans to be carried out.
   * @param pfirsts  The scanned values.
   * @param ptotals  The aux values (results of reductions).
   * @param pa       2D process arangement.
   * @param rs       2D rank.
   * @param genindex An indexer to generate 2D ranks of targets.
   */
  template <typename A, typename OP, typename INDEXER>
  void
  _impl::ms_adpt::global_scan_common(const OP& oplus,
                                     const int n,
                                     A* pfirsts,
                                     A* ptotals,
                                     const msize pa,
                                     const mindex rs,
                                     const INDEXER &genindex)
  {
    const int irank = genindex.rank(rs, pa);  // virtual rank in this direction
    const int iprocs = genindex.procs(rs, pa);// virtual procs of this direction
    A *received_vals = new A[n];
    bool is_firstval_set = false;
    for (int stage = 1; stage < iprocs; stage <<= 1) {
      const int itarget = irank ^ stage;
      if (itarget >= iprocs) continue;
      { // send-recv
        // the actual 2D rank of the target is given by genindex, and
        // its real rank is given by function compute_rank.
        int target = _impl::dist_matrix_buffer<A>::compute_rank(pa, genindex(itarget, rs, pa));  
        
        _impl::transporter::sendrecv<A>(ptotals, received_vals, n, target, TAG_SCAN1);
      }
      if (itarget > irank) { // perform n scans at once
        for(int i = 0; i < n; i++) {  
          ptotals[i] = oplus(ptotals[i], received_vals[i]);
        }
      } else {
        for(int i = 0; i < n; i++) {
          ptotals[i] = oplus(received_vals[i], ptotals[i]);
        }
        if (is_firstval_set) {
          for(int i = 0; i < n; i++) {
            pfirsts[i] = oplus(received_vals[i], pfirsts[i]);
          }
        } else {
          for(int i = 0; i < n; i++) {
            pfirsts[i] = received_vals[i]; 
          }
          is_firstval_set = true;
        }
      }
    }
    delete [] received_vals;
  }

  /*
   * This function performs scan according to the given indexers.
   *
   * We can make four scans by instantiating this function.
   *
   * @param bs       The destination matrix.
   * @param oplus    An associative operator used in reduce. This should 
   *                 be a function object of type "A oplus(A, A)".
   *                 This shuold have the abide property with otimes.
   * @param otimes   An associative operator used in reduce. This should
   *                 be a function object of type "A otimes(A, A)".
   *                 This shuold have the abide property with oplus.
   * @param as       The input matrix.
   * @param iindexer An indexer for i-direction
   * @param jindexer An indexer for j-direction
   *
   *
   */
  template <typename OPLUS, typename OTIMES, typename AS, typename A, typename IINDEXER, typename JINDEXER>
  dist_matrix<A>
  _impl::ms_adpt::scan_common(dist_matrix<A> *bs,
                              const OPLUS &oplus,
                              const OTIMES &otimes,
                              const AS &as,
                              const IINDEXER& iindexer,
                              const JINDEXER& jindexer)
  {  
    A* bs_buf = bs->data->buf;
    
    const msize local_size = as.get_local_size();
    const int ni = local_size.i;
    const int nj = local_size.j;
    const msize pa = _impl::dist_matrix_buffer<A>::compute_process_arrangement(as.get_global_size(), sketo::procs);
    mindex rs = _impl::dist_matrix_buffer<A>::compute_ranks(pa, sketo::rank);
    const int irank = rs.i;
    const int jrank = rs.j;
    const int ibegin = iindexer.begin(pa, local_size);
    const int iend = iindexer.end(pa, local_size);
    const int istep = iindexer.step(pa, local_size);
    const int iedgeproc = iindexer.endproc(pa, local_size);
    const int jbegin = jindexer.begin(pa, local_size);
    const int jend = jindexer.end(pa, local_size);
    const int jstep = jindexer.step(pa, local_size);
    const int jedgeproc = jindexer.endproc(pa, local_size);

    A *buf = new A[std::max(ni, nj)];
    A *buf2 = new A[std::max(ni, nj)];

    // local scan in j-direction
    {
      for(int i = ibegin; i != iend; i+=istep) {
        bs_buf[i * nj + jbegin] = as.local_get(mindex(i, jbegin));
        for(int j = jbegin + jstep; j != jend; j+=jstep) {
          bs_buf[i * nj + j] = otimes(bs_buf[i * nj + (j-jstep)], as.local_get(mindex(i, j)));
        }
      }
      for(int i = ibegin; i != iend; i+=istep) {
        buf[i] = bs_buf[i * nj + jend - jstep];
      }
    }

    // global_scan in j-direction
    global_scan_common(otimes, ni, buf2, buf, pa, rs, jindexer);

    // local_map in j-direction
    {
      if(jrank == jedgeproc) { /* do nothing */ }
      else {
        for(int i = ibegin; i != iend; i+=istep) {
          for(int j = jbegin; j != jend; j+=jstep) {
            bs_buf[i*nj + j] = otimes(buf2[i], bs_buf[i*nj + j]);
          }
        }
      }
    }

    // local scan in i-direction
    {
      for(int j = jbegin; j != jend; j+=jstep) {
        for(int i = ibegin + istep; i != iend; i+=istep) {
          bs_buf[i * nj + j] = oplus(bs_buf[(i - istep) * nj + j], bs_buf[i * nj + j]);
        }
      }
      for(int j = jbegin; j != jend; j+=jstep) {
        buf[j] = bs_buf[(iend - istep) * nj + j];
      }
    }
    
    // global_scan in i-direction
    global_scan_common(oplus, nj, buf2, buf, pa, rs, iindexer);
    
    // local_map in i-direction
    {
      if(irank == iedgeproc) { /* do nothing */ }
      else {
        for(int j = jbegin; j != jend; j+=jstep) {
          for(int i = ibegin; i != iend; i+=istep) {
            bs_buf[i*nj + j] = oplus(buf2[j], bs_buf[i*nj + j]);
          }
        }
      }
    }
 
    delete [] buf;
    delete [] buf2;
    return *bs;
  }

  /**
   * index information for various scans
   */
  struct scan_iindexer {  // the indexer for i-direction in scan.
    // the start index of the local scan
    int begin(msize pa, msize local_size) const { return 0; }
    // the end index of the local scan
    int end(msize pa, msize local_size) const { return local_size.i; }
    // the direction of the local scan
    int step(msize pa, msize local_size) const { return 1; }
    // the rank of the head process
    int endproc(msize pa, msize local_size) const { return 0; }
    // the virtual rank of this process during the one-direction global scan
    int rank(mindex rs, msize pa) const { return rs.i; }
    // the number of processes in this direction during the global scan
    int procs(mindex rs, msize pa) const { return pa.i; }
    // the actual 2D rank of the target process during the global scan
    mindex operator()(int target, mindex rs, msize pa) const {
      return mindex(target, rs.j);
    }
  };
  struct scan_jindexer {  // the indexer for j-direction in scan.
    int begin(msize pa, msize local_size) const { return 0; }
    int end(msize pa, msize local_size) const { return local_size.j; }
    int step(msize pa, msize local_size) const { return 1; }
    int endproc(msize pa, msize local_size) const { return 0; }
    int rank(mindex rs, msize pa) const { return rs.j; }
    int procs(mindex rs, msize pa) const { return pa.j; }
    mindex operator()(int target, mindex rs, msize pa) const {
      return mindex(rs.i, target);
    }
  };

  struct scanr_iindexer {  // the indexer for i-direction in rscan.
    int begin(msize pa, msize local_size) const { return local_size.i-1; }
    int end(msize pa, msize local_size) const { return -1; }
    int step(msize pa, msize local_size) const { return -1; }
    int endproc(msize pa, msize local_size) const { return pa.i-1; }
    int rank(mindex rs, msize pa) const { return pa.i - 1 - rs.i; }
    int procs(mindex rs, msize pa) const { return pa.i; }
    mindex operator()(int target, mindex rs, msize pa) const {
      return mindex(pa.i - 1 - target, rs.j);
    }
  };
  struct scanr_jindexer {  // the indexer for j-direction in rscan.
    int begin(msize pa, msize local_size) const { return local_size.j-1; }
    int end(msize pa, msize local_size) const { return -1; }
    int step(msize pa, msize local_size) const { return -1; }
    int endproc(msize pa, msize local_size) const { return pa.j-1; }
    int rank(mindex rs, msize pa) const { return pa.j - 1 - rs.j; }
    int procs(mindex rs, msize pa) const { return pa.j; }
    mindex operator()(int target, mindex rs, msize pa) const {
      return mindex(rs.i, pa.j - 1 - target);
    }
  };

  
// swaps arguments of the given associative operator
  template <typename OP, typename A>
  struct swap_op : public sketo::functions::base <A(A,A)> {
    const OP &op;
    swap_op(const OP &op) : op(op) { }
    A operator()(const A &a, const A&b) const { return op(b, a); }
  };
  
  template <typename OPLUS, typename OTIMES, typename A, typename AS>
  dist_matrix<A>
  _impl::ms_adpt::scan(dist_matrix<A> *bs,
                       const OPLUS &oplus,
                       const OTIMES &otimes,
                       const AS &as)
  {
    return scan_common(bs, oplus, otimes, as, scan_iindexer(), scan_jindexer());
  }

  template <typename OPLUS, typename OTIMES, typename A, typename AS>
  dist_matrix<A>
  _impl::ms_adpt::scanr(dist_matrix<A> *bs,
                        const OPLUS &oplus,
                        const OTIMES &otimes,
                        const AS &as)
  {
    return scan_common(bs, swap_op<OPLUS, A>(oplus), swap_op<OTIMES, A>(otimes), as,
                       scanr_iindexer(), scanr_jindexer());
  }
  template <typename AS>
  struct submatrix_serializer : sketo::serialization::require_serialization {
    const AS *as;
    const mindex begin;
    const msize size;
    submatrix_serializer(const AS *as, mindex begin, msize size) : as(as), begin(begin), size(size) { }
  private:
    //friend class boost::serialization::access;
    friend class sketo::serialization::access;
    
    template<class Archive>
    void serialize(Archive& ar, const unsigned int version)
      {
        for(int i = begin.i; i < begin.i + size.i; i++) {
          for(int j = begin.j; j < begin.j + size.j; j++) {
            ar & as->local_get(mindex(i, j)); // output only!
          }
        }
      }
  };

  template <typename A>
  struct submatrix_deserializer : sketo::serialization::require_serialization {
    dist_matrix<A> *as;
    const mindex begin;
    const msize size;
    A *buf;
    submatrix_deserializer(dist_matrix<A> *as, mindex begin, msize size) : as(as), begin(begin), size(size), buf(NULL) { }
    submatrix_deserializer(A* buf, msize size) : as(NULL), begin(mindex(0,0)), size(size), buf(buf) { }
  private:
    //friend class boost::serialization::access;
    friend class sketo::serialization::access;
    
    template<class Archive>
    void serialize(Archive& ar, const unsigned int version) {
      if(buf==NULL) {
        for(int i = begin.i; i < begin.i + size.i; i++) {
          for(int j = begin.j; j < begin.j + size.j; j++) {
            A tmp;
            ar & tmp;
            as->local_set(mindex(i, j), tmp);
          }
        }
      } else {
        for(int i = 0; i < size.i; i++) {
          for(int j = 0; j < size.j; j++) {
            ar & buf[i * size.j + j];
          }
        }
      }
    }
  };  
  /** @brief Common implementation for shifts.
   * @param bs  Target buffer to which the shifted results will be stored.
   *            If it is NULL, then the internal buffers of si is used.
   * @param si  Information about shift operation.
   * @param f   The function to be applied to elements outside the original space.
   * @param as  The original data.
   */ 
  template <int di, int dj, typename F, typename AS, typename A>
  void
  _impl::ms_adpt::shift_common(dist_matrix<A> *bs, shift_info<di, dj, A> &si, const F&f, const AS&as) {

    // transfer data
    simple_send_request< submatrix_serializer<AS> > reqs[4];
    simple_recv_request< submatrix_deserializer< A > > reqr[4];
    
    submatrix_deserializer<A> *sds[4];

    bool waitings[4];
    const msize global_size = as.get_global_size();
    
    const msize pa = _impl::dist_matrix_buffer<A>::compute_process_arrangement(global_size, sketo::procs);
    
    // sending data
    for(int i = 0; i < 4; i++) {
      if(si.need_send_communication(i)) {
        waitings[i] = true;

        int target = _impl::dist_matrix_buffer<A>::compute_rank(pa, si.to_sbis[i].rank);
        submatrix_serializer<AS> ss(&as, si.to_sbis[i].begin, si.to_sbis[i].size);
        _impl::transporter::isend(&ss, 1, target, TAG_SHIFT+si.to_sbis[i].tag, 
                                  &reqs[i]);
      } else {
        waitings[i] = false;
      }
    }

    // receiving data
    for(int i = 0; i < 4; i++) {
      if(si.need_receive_communication(i)) {
        int target = _impl::dist_matrix_buffer<A>::compute_rank(pa, si.from_sbis[i].rank);
        const int ni = si.from_sbis[i].size.i;
        const int nj = si.from_sbis[i].size.j;
        if(bs==NULL) {
          A*buf = si.buf->reserve(i, ni * nj);
          sds[i] = new submatrix_deserializer<A>(buf, si.from_sbis[i].size);
        } else {
          sds[i] = new submatrix_deserializer<A>(bs, si.from_sbis[i].begin, si.from_sbis[i].size);
        }
        _impl::transporter::irecv(sds[i], 1, target, TAG_SHIFT+i, 
                                  &reqr[i]);
      } else {
        sds[i] = NULL;
      }
    }
    

    for(int i = 0; i < 4; i++) {
      // self copy for the region outside
      if(si.need_receive_buf(i) && !si.need_receive_communication(i)) {  
        const int ni = si.from_sbis[i].size.i;
        const int nj = si.from_sbis[i].size.j;
        const int gi = si.from_sbis[i].global_begin.i;
        const int gj = si.from_sbis[i].global_begin.j;
        const int ti = si.from_sbis[i].target_begin.i;
        const int tj = si.from_sbis[i].target_begin.j;
        A*buf = si.buf->reserve(i, ni * nj);
        for(int i = 0; i < ni; i++) {
          for(int j = 0; j < nj; j++) {
            buf[i*nj + j] = f(mindex(gi + i, gj + j), as.local_get(mindex(ti + i, tj + j)));
          }
        }
      }
    }

    // deserialization
    if(bs == NULL) {
      for(int i = 0; i < 4; i++) {
        if(sds[i] != NULL) { //si.need_receive_communication(i)) {
          reqr[i].wait();
          A*buf = sds[i]->buf;
          
          const int ni = si.from_sbis[i].size.i;
          const int nj = si.from_sbis[i].size.j;
          const int gi = si.from_sbis[i].global_begin.i;
          const int gj = si.from_sbis[i].global_begin.j;
          if(si.from_sbis[i].from_outside) {
            for(int i = 0; i < ni; i++) {
              for(int j = 0; j < nj; j++) {
                buf[i*nj + j] = f(mindex(gi + i, gj + j), buf[i*nj + j]);
              }
            }
          }
        }
      }
    } else {
      // be carefull about the overwriting case that bs == as.
      for(int i = 0; i < 4; i++) { // self move
        if(si.is_self_move(i)) {
          const int bi = si.from_sbis[i].begin.i;
          const int bj = si.from_sbis[i].begin.j;
          const int ti = si.from_sbis[i].target_begin.i;
          const int tj = si.from_sbis[i].target_begin.j;
          const int ni = si.from_sbis[i].size.i;
          const int nj = si.from_sbis[i].size.j;
          if(bi < ti || (bi == ti && bj <= tj)) {
            // forward direction copying for col-major
            for(int i = 0; i < ni; i++) {
              for(int j = 0; j < nj; j++) {
                bs->local_set(mindex(bi + i, bj + j), as.local_get(mindex(ti + i, tj + j)));
              }
            }
            
          } else { 
            // backward direction copying for col-major
            for(int i = ni - 1; i >= 0; i--) {
              for(int j = nj - 1; j >= 0; j--) {
                bs->local_set(mindex(bi + i, bj + j), as.local_get(mindex(ti + i, tj + j)));
              }
            }
          }
        }
      }

      for(int i = 0; i < 4; i++) { // direct deserialization or copy
        const int ni = si.from_sbis[i].size.i;
        const int nj = si.from_sbis[i].size.j;
        const int gi = si.from_sbis[i].global_begin.i;
        const int gj = si.from_sbis[i].global_begin.j;
        const int bi = si.from_sbis[i].begin.i;
        const int bj = si.from_sbis[i].begin.j;
        if(sds[i] != NULL) { //si.need_receive_communication(i)) {
          reqr[i].wait();
          
          if(si.from_sbis[i].from_outside) {
            for(int i = 0; i < ni; i++) {
              for(int j = 0; j < nj; j++) {
                bs->local_set(mindex(bi + i, bj + j), f(mindex(gi + i, gj + j), bs->local_get(mindex(bi + i, bj + j))));
              }
            }
          }
        } else if(!si.is_self_move(i)){
          A*buf = si.buf->get_buf(i);
          for(int i = 0; i < ni; i++) {
            for(int j = 0; j < nj; j++) {
              bs->local_set(mindex(bi + i, bj + j), buf[i*nj + j]);
            }
          }
          
        }
      }
    }

    // wait for sending
    for(int i = 0; i < 4; i++) {
      if(waitings[i]) {
        reqs[i].wait();
      }
    }
  }
}

// namespace sketo {
//   namespace serialization {
//     template <typename A>
//     struct serialization_trait< submatrix_serializer<A> > {
//       static const bool need_serialization = true;   // tobe modified.
//     };
//     template <typename A>
//     struct serialization_trait< submatrix_deserializer<A> > {
//       static const bool need_serialization = true;   // tobe modified.
//     };
//   }
// }


#endif // __MATRIX_SKELETONS_COMMON_TPP__

