/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2010, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/

/**
 * @file matrix_skeletons_noopt.h
 * @brief An implmentation of matrix skeletons without fusion optimization.
 *        These skeletons are for debugging.
 */

#ifndef __MATRIX_SKELETONS_NOOPT_H__
#define __MATRIX_SKELETONS_NOOPT_H__

namespace sketo {
  namespace matrix_skeletons {
    /**
     * generate :: msize -> (mindex -> a) -> [a]
     */
    template <typename F>
    dist_matrix<typename result_of<F(mindex)>::type>
    generate(msize global_size,
             const F& f);

    /**
     * map :: (a -> b) -> [a] -> [b]
     */
    template <typename F, typename A>
    dist_matrix<typename result_of<F(A)>::type>
    map(F f,
        const sketo::dist_matrix<A>& as);

    /**
     * map_with_index :: (mindex -> a -> b) -> [a] -> [b]
     */
    template <typename F, typename A>
    dist_matrix<typename result_of<F(mindex, A)>::type>
    map_with_index(const F& f,
                   const dist_matrix<A>& as);

    /**
     * zip :: [a] -> [b] -> [(a, b)]
     */
    template <typename A, typename B>
    dist_matrix<std::pair<A, B> >
    zip(const dist_matrix<A>& as,
        const dist_matrix<B>& bs);

    /**
     * zipwith :: (a -> b -> c) -> [a] -> [b] -> [c]
     */
    template <typename F, typename A, typename B>
    dist_matrix<typename result_of<F(A, B)>::type>
    zipwith(const F& f,
            const dist_matrix<A>& as,
            const dist_matrix<B>& bs);

    /**
     * reduce :: (a -> a -> a) -> (a -> a -> a) -> [a] -> [a]
     */
    template <typename OPLUS, typename OTIMES, typename A>
    A
    reduce(const OPLUS &oplus,
           const OTIMES &otimes,
           const dist_matrix<A> &as) {
      return _impl::ms_impl_select::ReduceSimple::reduce<OPLUS, OTIMES, dist_matrix<A>, A>(oplus, otimes, as);
    }

    /**
     * reduce :: (a -> a -> a) -> (a -> a -> a) -> [a] -> [a]
     */
    template <typename OPLUS, typename OTIMES, typename IMPL, typename A>
    A
    reduce(const OPLUS &oplus,
           const OTIMES &otimes,
           const dist_matrix<A> &as,
           const IMPL &) {
      return IMPL::template reduce<OPLUS, OTIMES, dist_matrix<A>, A>(oplus, otimes, as);
    }
    
    /**
     * scan :: (a -> a -> a) -> (a -> a -> a) -> [a] -> [a]
     */
    template <typename OPLUS, typename OTIMES, typename A>
    dist_matrix<A>
    scan(const OPLUS &oplus,
         const OTIMES &otimes,
         const dist_matrix<A>& as);

    /**
     * scanr :: (a -> a -> a) -> (a -> a -> a) -> [a] -> [a]
     */
    template <typename OPLUS, typename OTIMES, typename A>
    dist_matrix<A>
    scanr(const OPLUS &oplus,
          const OTIMES &otimes,
          const dist_matrix<A>& as);

    /**
     * shift :: (Int, Int) -> (((Int, Int), a) -> a) -> [a] -> [a]
     */
    template <int di, int dj, typename F, typename A>
    dist_matrix<A>
    shift(const F &f,
          const dist_matrix<A>& as);
    
    /**
     * @todo write comments
     */
    template <typename A>
    void
    print(const dist_matrix<A>& as,
          std::ostream &stream = std::cout);
    
  }
}

#include "matrix_skeletons_noopt.tpp"

#endif //__MATRIX_SKELETONS_WITHOUT_FUSION_H__
