/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2010, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/

#ifndef __TRANSPORTER_H__
#define __TRANSPORTER_H__

#include "serialization.h"

namespace sketo {
  namespace _impl {
    enum {
      TAG_ALL_GATHER = 1001,
    };
    /** structure to hold serialized bufferes for Isend */
    template <typename A>
    struct simple_send_request {
      MPI_Request req;
      sketo::serialization::binary_oarchive oa;
      void wait() {
        MPI_Wait(&req, MPI_STATUS_IGNORE);
      }
      simple_send_request() { }
    };
      
    /** structure to hold serialized bufferes for Irecv */
    template <typename A>
    struct simple_recv_request {
      MPI_Request req;
      unsigned char *buf;
      int size;
      sketo::serialization::region_serializer<A> s;
      void wait() {
        MPI_Wait(&req, MPI_STATUS_IGNORE);
        if(buf!=NULL) {
          sketo::serialization::binary_iarchive ia(buf, size);
          ia >> s;
          delete buf;
        }
        buf = NULL;
      }
      simple_recv_request() : buf(NULL) { }
    };

    /*
     * This struct provides a set of functions for data transfer. It selects
     * concrete transfer functions according to the REQUEST_SERIALIZATION.
     * see serialization.h
     */
    struct transporter {
      template <typename A> 
      static void
      send(const A *p, int n, int target, int tag, MPI_Comm comm = MPI_COMM_WORLD);

      template <typename A>
      static void
      recv(A *p, int n, int target, int tag, MPI_Comm comm = MPI_COMM_WORLD);

      template <typename A>
      static void
      bcast(A *p, int n, int root, MPI_Comm comm = MPI_COMM_WORLD);

      template <typename A>
      static void
      sendrecv(const A *src, A *dst, int n, int target, int tag, MPI_Comm comm = MPI_COMM_WORLD);

      /** NOTE: The serialized data body is sent by Isend, while the size of serialized data is sent by Send */
      template <typename A>
      static void
      isend(const A *p, int n, int target, int tag, simple_send_request<A> *req, MPI_Comm comm = MPI_COMM_WORLD);

      template <typename A>
      static void
      irecv(A *p, int n, int target, int tag, simple_recv_request<A> *req, MPI_Comm comm = MPI_COMM_WORLD);

      template <typename A>
      static void
      allgather(A *items, int n, A* all, MPI_Comm comm = MPI_COMM_WORLD);
      
      static void
      barrier(MPI_Comm comm = MPI_COMM_WORLD);
    };
 
    /*
     * This struct provides a set of functions for data transfer with
     * serialization among processes. The type A should implement
     * the following serialization function and 'friend' declaration.

     private:
     template<class Archive>
     void serialize(Archive& ar, const unsigned int version);

     friend class sketo::serialization::access;
         
    */
    struct simple_serialization_transporter {
      template <typename A> 
      static void
      send(const A *p, int n, int target, int tag, MPI_Comm comm = MPI_COMM_WORLD);

      template <typename A>
      static void
      recv(A *p, int n, int target, int tag, MPI_Comm comm = MPI_COMM_WORLD);

      template <typename A>
      static void
      bcast(A *p, int n, int root, MPI_Comm comm = MPI_COMM_WORLD);

      template <typename A>
      static void
      sendrecv(const A *src, A *dst, int n, int target, int tag, MPI_Comm comm = MPI_COMM_WORLD);

      /** NOTE: The serialized data body is sent by Isend, while the size of serialized data is sent by Send */
      template <typename A>
      static void
      isend(const A *p, int n, int target, int tag, simple_send_request<A> *req, MPI_Comm comm = MPI_COMM_WORLD);

      template <typename A>
      static void
      irecv(A *p, int n, int target, int tag, simple_recv_request<A> *req, MPI_Comm comm = MPI_COMM_WORLD);

      template <typename A>
      static void
      allgather(A *items, int n, A* all, MPI_Comm comm = MPI_COMM_WORLD);
      
      static void
      barrier(MPI_Comm comm = MPI_COMM_WORLD);
    };

    /*
     * This struct provides a set of functions for sipmle solid data transfer.
     * The type A should be solid on the memory space (i.e., any instance ia of
     * A is completely specified by the memory ((char*)(&ia))[sizeof(A)].
    */
    struct simple_solid_transporter {
      template <typename A> 
      static void
      send(const A *p, int n, int target, int tag, MPI_Comm comm = MPI_COMM_WORLD);

      template <typename A>
      static void
      recv(A *p, int n, int target, int tag, MPI_Comm comm = MPI_COMM_WORLD);

      template <typename A>
      static void
      bcast(A *p, int n, int root, MPI_Comm comm = MPI_COMM_WORLD);

      template <typename A>
      static void
      sendrecv(const A *src, A *dst, int n, int target, int tag, MPI_Comm comm = MPI_COMM_WORLD);

      /** NOTE: The serialized data body is sent by Isend, while the size of serialized data is sent by Send */
      template <typename A>
      static void
      isend(const A *p, int n, int target, int tag, simple_send_request<A> *req, MPI_Comm comm = MPI_COMM_WORLD);

      template <typename A>
      static void
      irecv(A *p, int n, int target, int tag, simple_recv_request<A> *req, MPI_Comm comm = MPI_COMM_WORLD);

      template <typename A>
      static void
      allgather(A *items, int n, A* all, MPI_Comm comm = MPI_COMM_WORLD);
      
      static void
      barrier(MPI_Comm comm = MPI_COMM_WORLD);
    };

    template <bool x, typename T, typename F>
    struct select {
      typedef F value;
    };
    template <typename T, typename F>
    struct select<true, T, F> {
      typedef T value;
    };
    /**
     * The structure to select the concrete transporter.
     */
    template <typename A>
    struct transporter_selector {
      typedef typename select<sketo::serialization::serialization_trait<A>::need_serialization, simple_serialization_transporter, simple_solid_transporter>::value transporter;
    };

  }
}

#endif //__TRANSPORTER_H__
#ifndef __TRANSPORTER_TPP__
#define __TRANSPORTER_TPP__
namespace sketo { 
  template <typename A>
  void
  _impl::simple_serialization_transporter::send(const A *p, int n, int target, int tag,
                                                MPI_Comm comm) {

    sketo::serialization::region_serializer<A> s(p, 0, n, 0, 1);
    sketo::serialization::binary_oarchive oa;
    oa << s;
    unsigned int size = oa.size();
    MPI_Send(&size, sizeof(unsigned int), MPI_BYTE, target, tag, comm);
    MPI_Send((void*)oa.buf(), size, MPI_BYTE, target, tag, comm);
  }

  template <typename A>
  void
  _impl::simple_serialization_transporter::recv(A *p, int n, int target, int tag,
                                                MPI_Comm comm) {
    unsigned int size = 0;
    MPI_Recv(&size, sizeof(unsigned int), MPI_BYTE, target, tag, comm,
             MPI_STATUS_IGNORE);
    
    unsigned char *buf = new unsigned char[size];
    MPI_Recv(buf, size, MPI_BYTE, target, tag, comm, MPI_STATUS_IGNORE);
    sketo::serialization::binary_iarchive ia(buf, size);
    sketo::serialization::region_serializer<A> s(p, 0, n, 0, 1);
    ia >> s;
    
    delete buf;
  }

  template <typename A>
  void
  _impl::simple_serialization_transporter::bcast(A *p, int n, int root,
                                                 MPI_Comm comm) {
    const int tag = 15;

    int procs = sketo::procs;
    int myrank = sketo::rank;
    int rank = (myrank - root + procs) % procs;
    int stage = 1;
    for(; stage < procs; stage <<= 1);
    for (stage >>= 1; stage > 0; stage >>= 1) { // tree-style communication
      const int target = rank ^ stage;
      if (target >= procs) continue;
      if (target > rank) {
        int mytarget = (target + root) % procs;
        _impl::simple_serialization_transporter::send(p, n, mytarget, tag, comm);
      } else {
        int mytarget = (target + root) % procs;
        _impl::simple_serialization_transporter::recv(p, n, mytarget, tag, comm);
      }
    }
  }
  template <typename A>
  void
  _impl::simple_serialization_transporter::sendrecv(const A *src, A *dst, int n, int target, int tag, MPI_Comm comm) {

    sketo::serialization::region_serializer<A> ssrc(src, 0, n, 0, 1);
    sketo::serialization::region_serializer<A> sdst(dst, 0, n, 0, 1);

    sketo::serialization::binary_oarchive oa;
    
    oa << ssrc;
    unsigned int size = oa.size();
    unsigned int rsize = 0;

    MPI_Send(&size, sizeof(unsigned int), MPI_BYTE, target, tag, comm);

    MPI_Recv(&rsize, sizeof(unsigned int), MPI_BYTE, target, tag, comm,
             MPI_STATUS_IGNORE);

    MPI_Request reqs, reqr;
    unsigned char *buf = new unsigned char[rsize];
    MPI_Isend((void*)oa.buf(), size, MPI_BYTE, target, tag, comm, &reqs);
    MPI_Irecv(buf, rsize, MPI_BYTE, target, tag, comm, &reqr);
    MPI_Wait(&reqr, MPI_STATUS_IGNORE);
    
    sketo::serialization::binary_iarchive ia(buf, rsize);
    ia >> sdst;
    
    MPI_Wait(&reqs, MPI_STATUS_IGNORE);
    
    delete buf;
  }
  
  template <typename A>
  void
  _impl::simple_serialization_transporter::isend(const A *p, int n, int target, int tag,
                                                 _impl::simple_send_request<A> *req,
                                                 MPI_Comm comm) {

    sketo::serialization::region_serializer<A> s(p, 0, n, 0, 1);
    req->oa << s;
    unsigned int size = req->oa.size();
    MPI_Send(&size, sizeof(unsigned int), MPI_BYTE, target, tag, comm);
    MPI_Isend((void*)req->oa.buf(), size, MPI_BYTE, target, tag, comm, &(req->req));
  }
  template <typename A>
  void
  _impl::simple_serialization_transporter::irecv(A *p, int n, int target, int tag,
                                                 _impl::simple_recv_request<A> *req,
                                                 MPI_Comm comm) {
    MPI_Recv(&(req->size), sizeof(unsigned int), MPI_BYTE, target, tag, comm,
             MPI_STATUS_IGNORE);
    
    req->buf = new unsigned char[req->size];
    MPI_Irecv(req->buf, req->size, MPI_BYTE, target, tag, comm, &(req->req));
    sketo::serialization::region_serializer<A> s(p, 0, n, 0, 1);
    req->s = s;
  }
  void
  _impl::simple_serialization_transporter::barrier(MPI_Comm comm) {
    MPI_Barrier(comm);
  }
  template <typename A>
  void
  _impl::simple_serialization_transporter::allgather(A *items, int n, A *all, MPI_Comm comm) {
    const int procs = sketo::procs;
    const int rank = sketo::rank;
    for(int p = 0; p < procs; p++) {
      if(p==rank) {
        sketo::serialization::binary_oarchive oa;
        sketo::serialization::region_serializer<A> rs(items, 0, n, 0, 1);
        oa << rs;        
        for(int pp = 0; pp < procs; pp++) {
          if(p==pp) continue;
          unsigned int size = oa.size();
          MPI_Send(&size, sizeof(unsigned int), MPI_BYTE, pp, TAG_ALL_GATHER, comm);
          MPI_Send((void*)oa.buf(), size, MPI_BYTE, pp, TAG_ALL_GATHER, comm);
        }
      } else {
        recv((all + n * p), n, p, TAG_ALL_GATHER);
      }
    }
  }
}

namespace sketo { 
  template <typename A>
  void
  _impl::simple_solid_transporter::send(const A *p, int n, int target, int tag,
                                                MPI_Comm comm) {
    MPI_Send((void*)p, n * sizeof(A), MPI_BYTE, target, tag, comm);
  }

  template <typename A>
  void
  _impl::simple_solid_transporter::recv(A *p, int n, int target, int tag,
                                                MPI_Comm comm) {
    MPI_Recv((void*)p, n * sizeof(A), MPI_BYTE, target, tag, comm, MPI_STATUS_IGNORE);
  }

  template <typename A>
  void
  _impl::simple_solid_transporter::bcast(A *p, int n, int root,
                                                 MPI_Comm comm) {
    MPI_Bcast((void*)p, sizeof(A) * n, MPI_BYTE, root, comm);
  }
  
  template <typename A>
  void
  _impl::simple_solid_transporter::sendrecv(const A *src, A *dst, int n, int target, int tag, MPI_Comm comm) {

    MPI_Sendrecv((void*)src, n * sizeof(A), MPI_BYTE, target, tag,
                 (void*)dst, n * sizeof(A), MPI_BYTE, target, tag,
                 comm, MPI_STATUS_IGNORE);
  }
  
  template <typename A>
  void
  _impl::simple_solid_transporter::isend(const A *p, int n, int target, int tag,
                                         _impl::simple_send_request<A> *req,
                                         MPI_Comm comm) {
    MPI_Isend((void*)p, n * sizeof(A), MPI_BYTE, target, tag, comm, &(req->req));
  }
  
  template <typename A>
  void
  _impl::simple_solid_transporter::irecv(A *p, int n, int target, int tag,
                                                 _impl::simple_recv_request<A> *req,
                                                 MPI_Comm comm) {
    MPI_Irecv((void*)p, n* sizeof(A), MPI_BYTE, target, tag, comm, &(req->req));
  }
  void
  _impl::simple_solid_transporter::barrier(MPI_Comm comm) {
    MPI_Barrier(comm);
  }
  template <typename A>
  void
  _impl::simple_solid_transporter::allgather(A *items, int n, A *all, MPI_Comm comm) {
    MPI_Allgather((void*)items, n * sizeof(A), MPI_BYTE,
                  (void*)all, n * sizeof(A), MPI_BYTE, comm);
  }
}

namespace sketo { 
  template <typename A>
  void
  _impl::transporter::send(const A *p, int n, int target, int tag,
                                                MPI_Comm comm) {
    transporter_selector<A>::transporter::send(p, n, target, tag, comm);
  }

  template <typename A>
  void
  _impl::transporter::recv(A *p, int n, int target, int tag,
                                                MPI_Comm comm) {
    transporter_selector<A>::transporter::recv(p, n, target, tag, comm);
  }

  template <typename A>
  void
  _impl::transporter::bcast(A *p, int n, int root,
                                                 MPI_Comm comm) {
    transporter_selector<A>::transporter::bcast(p, n, root, comm);
  }
  
  template <typename A>
  void
  _impl::transporter::sendrecv(const A *src, A *dst, int n, int target, int tag, MPI_Comm comm) {

    transporter_selector<A>::transporter::sendrecv(src, dst, n, target, tag, comm);
  }
  
  template <typename A>
  void
  _impl::transporter::isend(const A *p, int n, int target, int tag,
                                         _impl::simple_send_request<A> *req,
                                         MPI_Comm comm) {
    transporter_selector<A>::transporter::isend(p, n, target, tag, req, comm);
  }
  
  template <typename A>
  void
  _impl::transporter::irecv(A *p, int n, int target, int tag,
                                                 _impl::simple_recv_request<A> *req,
                                                 MPI_Comm comm) {
    transporter_selector<A>::transporter::irecv(p, n, target, tag, req, comm);
  }
  void
  _impl::transporter::barrier(MPI_Comm comm) {
    MPI_Barrier(comm);
    //transporter_selector<A>::transporter::barrier(comm);
  }
  template <typename A>
  void
  _impl::transporter::allgather(A *items, int n, A *all, MPI_Comm comm) {
    transporter_selector<A>::transporter::allgather(items, n, all, comm);
  }
}

#endif //__TRANSPORTER_TPP__

