/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2010, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/

/**
 * @file bracket_matching.cpp
 * @brief This program check bracket matching.
 * @author Yoshiki AKASHI (yoshiki@ipl.cs.uec.ac.jp)
 * @attention: If brackets does not match, this program may freeze.
 */

#include <sketo/functions.h>
#include <sketo/list_skeletons.h>
#include <iostream>
#include <cstdlib>
#include <ctime>
#include <stack>

const int KIND = 4;
const int DEPTH = 10;
const int MAXDEPTH = 10;
const char openbr[] = "({[<";
const char closebr[] = ")}]>";

//---------------------------------------------------------------------
// Stack
class Stack {
public:
  char in;
  char out;
  char data[ MAXDEPTH ];
  Stack( ): in( 0 ), out( 0 ) { }
  Stack( const Stack &s ): in( s.in ), out( s.out ) {
    for ( int i = 0; i < in; i++ ) {
      data[ i ] = s.data[ i ];
    }
  } 
  void push( char c ) {
    data[ in++ ] = c;
  }
  void pop( ){
    ( in > 0 ) ? in-- :  out++;
  }
  char top( ) const {
    return ( in > 0 ) ? data[ in-1 ] : 0;
  }
  int isempty( ) const {
    return ( in == 0 ) && ( out == 0 );
  }
  Stack& operator=( const Stack &s ) {
    if ( this != &s ) {
      in = s.in;
      out = s.out;
      for ( int i = 0; i < in; i++ ) {
	data[ i ] = s.data[ i ];
      }
    }
    return *this;
  }
};

//---------------------------------------------------------------------
// function declaration
int ran( int n );
void makestr( char *array, int size, int kind, int maxdepth );
int isopen( char c );
int isclose( char c );

//---------------------------------------------------------------------
// function definition
inline int ran( int n ) {
  return rand( ) % n;
}

inline void makestr( char *array, int size, int kind, int maxdepth ) {
  int n = 0;
  int *s = new int[ maxdepth+1 ];

  srand( time( NULL ) );

  while ( size > ( n + 2 ) ) {
    switch ( ( n == 0 ) ? ran( 2 ) : ran( 3 ) ) {
    case 0: // open
      if ( n >= maxdepth ) {
	continue;
      }
      *s = ran( kind );
      *array++ = openbr[ *s++ ];
      n++;
      break;
    case 1:
      *array++ = 'X';
      break;
    case 2: // close
      *array++ = closebr[ *--s ];
      n--;
      break;
    }
    size--;
  }
  while ( n-- > 0 ) {
    *array++ = closebr[ *--s ];
    size--;
  }
  while ( size-- > 1 ) {
    *array++ = 'X';
  }
  *array = '\0';
  delete [] s;
}

inline int isopen( char c ) {
  return ( c == '(' ) || ( c == '{' ) || ( c == '[' ) || ( c == '<' );
}

inline int isclose( char c ) {
  return ( c == ')' ) || ( c == '}' ) || ( c == ']' ) || ( c == '>' );
}

inline int match( char a, char b ) {
  if ( a == ')' ) return ( b == '(' );
  if ( a == '}' ) return ( b == '{' );
  if ( a == ']' ) return ( b == '[' );
  if ( a == '>' ) return ( b == '<' );
  return 0;
}

struct G : public sketo::functions::base<int(Stack)> {
  int operator()( const Stack &s ) const {
    return s.isempty( );
  }
} g;

struct P : public sketo::functions::base<int(char, Stack)> {
  int operator()( const char c, const Stack &s ) const {
    return isclose( c ) ? ( !s.isempty( ) && match( c, s.top( ) ) ) : 1;
  }
} p;

struct Oplus : public sketo::functions::base< int(int, int) > {
  int operator()( const int x, const int y ) const {
    return x && y;
  }
} oplus;

struct Q : public sketo::functions::base< Stack(char) > {
  Stack operator()( const char c ) const {
    Stack ret;
    if ( isopen( c ) ) {
      ret.push(c);
    }
    else if ( isclose( c ) ) {
      ret.pop( );
    }
    return ret;
  }
} q;

struct Otimes : public sketo::functions::base< Stack(Stack, Stack) > {
  Stack operator()( const Stack a, const Stack b ) const {
    if ( a.in <= b.out ) {
      Stack c(b);
      c.out += a.out - a.in;
      return c;
    }
    else {
      Stack c(a);
      for ( int i = 0; i < b.in; i++ ) {
	c.data[ a.in - b.out + i ] = b.data[ i ];
      }
      c.in += b.in - b.out;
      return c;
    }
  }
} otimes;

Stack identity_element(Otimes) {
  return Stack();
}


int sketo::main( int argc, char **argv )
{
  if ( argc < 3 ) {
    sketo::cout << "usage: " << argv[ 0 ] << " size loop" << std::endl;
    return 0;
  }

  const int size = std::atoi( argv[ 1 ] );
  const int loop = std::atoi( argv[ 2 ] );

//   sketo::time::initialize( ); // for time ekeeping

  // generate string include brackets
  char *str = new char[ size ];
  makestr( str, size, KIND, DEPTH );
  // output string
//   sketo::cout << str << std::endl;

  //   distribute data
  dist_list< char > as(size, str);

  double st_time = MPI_Wtime();
  
  for ( int i = 0; i < loop; i++ ) {
    // calculate bracket matching
    Stack st;
    int result;
//     list_skeletons::accumulate( g, p, oplus, q, otimes, &st, as, &result );

#ifndef __SLOW__
    result = sketo::list_skeletons::reduce(oplus,
sketo::list_skeletons::zipwith(p,
as,
sketo::list_skeletons::scan(otimes, Stack(),
sketo::list_skeletons::map(q, as), &st)));
    result = oplus(result, g(st));
    sketo::cout << ( ( result == 1 ) ? "match" : "no match" ) << std::endl;
#else
    dist_list<Stack> bs = sketo::list_skeletons::map(q, as);
    bs = sketo::list_skeletons::scan(otimes, Stack(), bs, &st); // overwrite
    dist_list<int> ds = sketo::list_skeletons::zipwith(p, as, bs);
    result = sketo::list_skeletons::reduce(oplus, ds);
    result = oplus(result, g(st));
#endif
//     // output result
//     sketo::cout << ( ( result == 1 ) ? "match" : "no match" ) << std::endl;
  }
  double ed_time = MPI_Wtime();
  sketo::cout << ed_time - st_time << std::endl;

  delete [] str;

//   sketo::time::set_time( "calculate" ); // for time ekeeping
//   sketo::time::print( ); // show the result of time ekeeping

  return 0;
}
