/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2010, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/

/**
 * @file test_list_skeletons.cpp
 * @brief Test code for list_skeletons.h and list_skeletons.cpp
 */

#include <mpi.h>
//#define __SKETO_NO_FUSION__

#include "../include/sketo/sketo.h"
#include "../include/sketo/list_skeletons.h"

#include <cassert>
#include <utility>
#include <iostream>
#include <functional>
#include <vector>
#include <string>
#include <cstdio>

const int size = 13;

typedef std::string string;
typedef std::pair<int, int> int_pair;
typedef std::pair<double, double> double_pair;
typedef std::vector<int> int_vector;
typedef std::vector<int_vector> int_mat;

template <typename A, typename B>
std::ostream& operator<<(std::ostream& os, const std::pair<A, B>& x) {
  os << "(" << x.first << "," << x.second << ")";
  return os;
}
template <typename A>
std::ostream& operator<<(std::ostream& os, const std::vector<A>& x) {
  os << "[";
  for(unsigned int i = 0; i < x.size(); i++) {
    if(i > 0) os << ",";
    os << x[i];
  }
  os << "]";
  return os;
}


template <typename T>
bool same_arrays(T* as, T* bs, int size)
{
  for (int i = 0; i < size; ++i) {
    if (as[i] != bs[i]) {
      std::cerr << "Wrong data: " << "LHS = " << as[i] << ", RHS = " << bs[i] << ". at index " << i << "." << std::endl;
      return false;
    }
  }
  return true;
}


class to_int_pair_t : public sketo::functions::base <int_pair(int)> {
public: int_pair operator()(int x) const { return int_pair(x + 1, x * x); }
} to_int_pair;

class to_int_vector_t : public sketo::functions::base <int_vector(int)> {
public:
  int_vector operator()(int ii) const {
    int_vector v;
    for(int i = 0; i <= ii; i++) {
      v.push_back(i);
    }
    return v;
  }
} to_int_vector;

class to_string_t : public sketo::functions::base <string(int)> {
public:
  string operator()(int ii) const {
    char buf[25];
    sprintf(buf, "%d", ii);
    string s(buf);
    return s;
  }
} to_string;

// function object for map_index
class vector_append_t : public sketo::functions::base <int_vector(int_vector, int_vector)> {
 public: int_vector operator()(const int_vector &x, const int_vector &y) const {
   int_vector im;
   for(unsigned int i = 0; i < x.size(); i++) im.push_back(x[i]);
   for(unsigned int i = 0; i < y.size(); i++) im.push_back(y[i]);  
   return im;
 }
} vector_append;

int_vector identity_element(const vector_append_t &)
{
  int_vector iv;
  return iv;
}

// function object for map
class same_last_two_bit_t : public sketo::functions::base <bool(int_pair)> {
 public: bool operator()(int_pair x) const { return ((x.first & 3) == (x.second & 3)); }
} same_last_two_bit;

// function object for map_index
class same_last_two_bit_with_index_t : public sketo::functions::base <bool(int, int_pair)> {
 public: bool operator()(int x, int_pair y) const { return (((x + y.first) & 3) == (y.second & 3)); }
} same_last_two_bit_with_index;

// function objects for zipwith
class to_double_pair_t : public sketo::functions::base <double_pair(int)> {
 public: double_pair operator()(int x) const { return double_pair(x + 1, x - 1); }
} to_double_pair;

class int_pair_with_doule_pair_t : public sketo::functions::base <bool(int_pair, double_pair)> {
 public: bool operator()(int_pair x, double_pair y) const { return ((double)x.first == y.first); };
} int_pair_with_double_pair;

// function objcets for gscanl and gscanr
class gscan_f_t : public sketo::functions::base <double (double, int)> {
 public: double operator()(double c, int x) const { return c - x * 2.0; }
} gscan_f;

class gscan_g_t : public sketo::functions::base <double (int)> {
 public: double operator()(int x) const { return x * 2.0; }
} gscan_g;

// function objects for accumulate
class accumulate_g_t : public sketo::functions::base <double_pair (double)> {
 public: double_pair operator()(double x) const { return double_pair(x, x); }
} accumulate_g;

class accumulate_p_t : public sketo::functions::base <double_pair (int, double)> {
public: double_pair operator()(int x, double y) const { return double_pair(-2 * x, y); }
} accumulate_p;

class accumulate_q_t : public sketo::functions::base <double (int)> {
 public: double operator()(int x) const { return -x; };
} accumulate_q;

class accumulate_oplus_t : public sketo::functions::base <double_pair (double_pair, double_pair)> {
 public: double_pair operator()(const double_pair &x, const double_pair &y) const {
   return double_pair(x.first + y.first, x.second + y.second);
 }
} accumulate_oplus;


void test_get_set()
{
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::dist_list<int_pair>* as = new sketo::dist_list<int_pair>(size);

  for (int i = 0; i < size; ++i) {
    as->set(i, to_int_pair(i));
  }

  for (int i = 0; i < size; ++i) {
    int_pair a = as->get(i);
    if (sketo::is_master()) {
      assert(a == to_int_pair(i));
    }
  }
  delete as;
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_get_set:OK" << std::endl;
}
void test_get_set2()
{
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::dist_list<int_vector>* as = new sketo::dist_list<int_vector>(size);

  for (int i = 0; i < size; ++i) {
    as->set(i, to_int_vector(i));
  }

  for (int i = 0; i < size; ++i) {
    int_vector a = as->get(i);
    if (sketo::is_master()) {
      assert(a == to_int_vector(i));
    }
  }
  delete as;
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_get_set2:OK" << std::endl;
}

void test_distribute_gather1()
{
  MPI_Barrier(MPI_COMM_WORLD);
  int_pair* array = new int_pair[size];
  for (int i = 0; i < size; ++i) {
    if (sketo::is_master()) array[i] = to_int_pair(i);
    else                    array[i] = to_int_pair(i+10);
  }
  
  sketo::dist_list<int_pair> as(size, array);
  for (int i = 0; i < size; ++i) {
    assert(as.get(i) == to_int_pair(i));
  }
  
  int_pair* array2 = new int_pair[size];
  as.gather(array2);
  if (sketo::is_master()) assert(same_arrays(array, array2, size));

  delete[] array;
  delete[] array2;

  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_distribute_gather1:OK" << std::endl;
}

void test_distribute_gather2()
{
  MPI_Barrier(MPI_COMM_WORLD);
  int_vector* array = new int_vector[size];
  for (int i = 0; i < size; ++i) {
    if (sketo::is_master()) array[i] = to_int_vector(i);
    else                    array[i] = to_int_vector(i+10);
  }
  
  sketo::dist_list<int_vector> as(size, array);
  for (int i = 0; i < size; ++i) {
    assert(as.get(i) == to_int_vector(i));
  }
  
  int_vector* array2 = new int_vector[size];
  as.gather(array2);
  if (sketo::is_master()) assert(same_arrays(array, array2, size));

  delete[] array;
  delete[] array2;

  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_distribute_gather2:OK" << std::endl;
}

void test_distribute_gather3()
{
  MPI_Barrier(MPI_COMM_WORLD);
  string* array = new string[size];
  for (int i = 0; i < size; ++i) {
    if (sketo::is_master()) array[i] = to_string(i);
    else                    array[i] = to_string(i+10);
  }
  
  sketo::dist_list<string> as(size, array);
  for (int i = 0; i < size; ++i) {
    assert(as.get(i) == to_string(i));
  }
  
  string* array2 = new string[size];
  as.gather(array2);
  if (sketo::is_master()) assert(same_arrays(array, array2, size));

  delete[] array;
  delete[] array2;

  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_distribute_gather3:OK" << std::endl;
}


void test_generate()
{
  MPI_Barrier(MPI_COMM_WORLD);
  // skeleton code
  sketo::dist_list<int_pair> as;
  as = sketo::list_skeletons::generate(size, to_int_pair);
  
  int_pair* array_skel = new int_pair[size];
  as.gather(array_skel);

  // sequential code
  int_pair* array_seq = new int_pair[size];
  for (int i = 0; i < size; ++i) {
    array_seq[i] = to_int_pair(i);
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_generate:OK" << std::endl;
}


void test_map()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  sketo::dist_list<bool> as;
  as = sketo::list_skeletons::map(same_last_two_bit,
				  sketo::list_skeletons::generate(size, to_int_pair));
  bool* array_skel = new bool[size];
  as.gather(array_skel);
  
  // sequential code
  bool* array_seq = new bool[size];
  for (int i = 0; i < size; ++i) {
    array_seq[i] = same_last_two_bit(to_int_pair(i));
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_map:OK" << std::endl;
}

void test_map_overwrite()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  sketo::dist_list<double> as;
  as = sketo::list_skeletons::generate(size, sketo::functions::caster<int, double>());
  as = sketo::list_skeletons::map(std::bind1st(std::multiplies<double>(), 2), as);
  double* array_skel = new double[size];
  as.gather(array_skel);

  // sequential code
  double *array_seq = new double[size];
  for (int i = 0; i < size; ++i) {
    array_seq[i] = 2 * static_cast<double>(i);
  }

  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;

  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_map_overwrite:OK" << std::endl;
}

void test_map_fusion()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  sketo::dist_list<double> as;
  as = sketo::list_skeletons::map(std::bind1st(std::multiplies<double>(), -1),
				  sketo::list_skeletons::map(std::bind1st(std::plus<double>(), -1),
							     sketo::list_skeletons::generate(size, sketo::functions::caster<int, double>())));
}


void test_map_with_index()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  sketo::dist_list<bool> bs;
  bs = sketo::list_skeletons::map_with_index(same_last_two_bit_with_index,
					     sketo::list_skeletons::generate(size, to_int_pair));
  bool* array_skel = new bool[size];
  bs.gather(array_skel);
  
  // sequential code
  bool* array_seq = new bool[size];
  for (int i = 0; i < size; ++i) {
    array_seq[i] = same_last_two_bit_with_index(i, to_int_pair(i));
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_map_with_index:OK" << std::endl;
}


void test_zip()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  std::pair<int_pair, double_pair> *array_skel = new std::pair<int_pair, double_pair>[size];
  {
    sketo::dist_list<std::pair <int_pair, double_pair> > cs;
    cs = sketo::list_skeletons::zip(sketo::list_skeletons::generate(size, to_int_pair),
				    sketo::list_skeletons::generate(size, to_double_pair));
    cs.gather(array_skel);
  }
      
  // sequential code
  std::pair <int_pair, double_pair> *array_seq = new std::pair <int_pair, double_pair>[size];
  for (int i = 0; i < size; ++i) {
    array_seq[i].first = to_int_pair(i);
    array_seq[i].second = to_double_pair(i);
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_zip:OK" << std::endl;
}


void test_zipwith()
{
  MPI_Barrier(MPI_COMM_WORLD);
  
  // skeleton code
  bool* array_skel = new bool[size];
  {
    sketo::dist_list<bool> cs;
    cs = sketo::list_skeletons::zipwith(int_pair_with_double_pair,
					sketo::list_skeletons::generate(size, to_int_pair),
					sketo::list_skeletons::generate(size, to_double_pair));
    cs.gather(array_skel);
  }

  // sequential code
  bool *array_seq = new bool[size];
  for (int i = 0; i < size; ++i) {
    array_seq[i] = int_pair_with_double_pair(to_int_pair(i), to_double_pair(i));
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_zipwith:OK" << std::endl;
}


void test_reduce()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int result_skel = sketo::list_skeletons::reduce(std::plus<int>(),
						  sketo::list_skeletons::generate(size, sketo::functions::identity<int>()));

  // sequential code
  int result_seq = 0;
  for (int i = 1; i < size; ++i) { result_seq += i; }

  // compare
  if (sketo::is_master()) assert(result_skel == result_seq);

  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_reduce:OK" << std::endl;
}


void test_scan()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int* array_skel = new int[size];
  int last_skel;
  {
    sketo::dist_list<int> as = sketo::list_skeletons::scan(std::plus<int>(),
							   3,
							   sketo::list_skeletons::generate(size, sketo::functions::identity<int>()),
							   &last_skel);
    as.gather(array_skel);
  }

  // sequential code
  int *array_seq = new int[size];
  int last_seq = 3;
  for (int i = 0; i < size; ++i) {
    array_seq[i] = last_seq; last_seq += i;
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));
  if (sketo::is_master()) assert(last_skel == last_seq);

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_scan:OK" << std::endl;
}


void test_scanr()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int* array_skel = new int[size];
  int first_skel;
  {
    sketo::dist_list<int> as = sketo::list_skeletons::scanr(std::plus<int>(),
							    3,
							    sketo::list_skeletons::generate(size, sketo::functions::identity<int>()),
							    &first_skel);
    as.gather(array_skel);
  }

  // sequential code
  int *array_seq = new int[size];
  int first_seq = 3;
  for (int i = size - 1; i >= 0; --i) {
    array_seq[i] = first_seq; first_seq += i;
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));
  if (sketo::is_master()) assert(first_skel == first_seq);

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_scanr:OK" << std::endl;
}

void test_reduce2()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int_vector result_skel = sketo::list_skeletons::reduce(vector_append,
                                                         sketo::list_skeletons::generate(size, to_int_vector));

  // sequential code
  int_vector result_seq;
  for (int i = 0; i < size; ++i) { result_seq = vector_append(result_seq, to_int_vector(i)); }

  // compare
  if (sketo::is_master()) assert(result_skel == result_seq);

  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_reduce2:OK" << std::endl;
}


void test_scan2()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int_vector* array_skel = new int_vector[size];
  int_vector last_skel;
  int_vector e;
  e.push_back(3);
  {
    sketo::dist_list<int_vector> as = sketo::list_skeletons::scan(vector_append,
							   e,
							   sketo::list_skeletons::generate(size, to_int_vector),
							   &last_skel);
    as.gather(array_skel);
  }

  // sequential code
  int_vector *array_seq = new int_vector[size];
  int_vector last_seq = e;
  for (int i = 0; i < size; ++i) {
    array_seq[i] = last_seq; last_seq = vector_append(last_seq, to_int_vector(i));
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));
  if (sketo::is_master()) assert(last_skel == last_seq);

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_scan2:OK" << std::endl;
}


void test_scanr2()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int_vector* array_skel = new int_vector[size];
  int_vector first_skel;
  int_vector e;
  e.push_back(3);
  {
    sketo::dist_list<int_vector> as = sketo::list_skeletons::scanr(vector_append,
							    e,
							    sketo::list_skeletons::generate(size, to_int_vector),
							    &first_skel);
    as.gather(array_skel);
  }

  // sequential code
  int_vector *array_seq = new int_vector[size];
  int_vector first_seq = e;
  for (int i = size - 1; i >= 0; --i) {
    array_seq[i] = first_seq; first_seq = vector_append(to_int_vector(i), first_seq);
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));
  if (sketo::is_master()) assert(first_skel == first_seq);

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_scanr2:OK" << std::endl;
}


void test_postscan()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int* array_skel = new int[size];
  {
    sketo::dist_list<int> as = sketo::list_skeletons::postscan(std::plus<int>(),
							       sketo::list_skeletons::generate(size, sketo::functions::identity<int>()));
    as.gather(array_skel);
  }

  // sequential code
  int *array_seq = new int[size];
  array_seq[0] = 0;
  for (int i = 1; i < size; ++i) {
    array_seq[i] = array_seq[i-1] + i;
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_postscan:OK" << std::endl;
}


void test_postscanr()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int* array_skel = new int[size];
  {
    sketo::dist_list<int> as = sketo::list_skeletons::postscanr(std::plus<int>(),
								sketo::list_skeletons::generate(size, sketo::functions::identity<int>()));
    as.gather(array_skel);
  }

  // sequential code
  int *array_seq = new int[size];
  array_seq[size - 1] = size - 1;
  for (int i = size - 1 - 1; i >= 0; --i) {
    array_seq[i] = array_seq[i+1] + i;
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_postscanr:OK" << std::endl;
}


void test_gscanl()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  double *array_skel = new double[size];
  {
    sketo::dist_list<int> as = sketo::list_skeletons::generate(size, sketo::functions::identity<int>());
    sketo::dist_list<double> bs = sketo::list_skeletons::gscanl(gscan_f, 100.0, gscan_g, std::minus<double>(), std::plus<double>(), as);
    bs.gather(array_skel);
  }

  // sequential code
  double *array_seq = new double[size];
  double val = 100;
  for (int i = 0; i < size; ++i) {
    array_seq[i] = val;
    val = gscan_f(val, i);
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;

  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_gscanl:OK" << std::endl;
}    
		       

void test_gscanr()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  double *array_skel = new double[size];
  {
    sketo::dist_list<int> as = sketo::list_skeletons::generate(size, sketo::functions::identity<int>());
    sketo::dist_list<double> bs = sketo::list_skeletons::gscanr(gscan_f, 100.0, gscan_g, std::minus<double>(), std::plus<double>(), as);
    bs.gather(array_skel);
  }

  // sequential code
  double *array_seq = new double[size];
  double val = 100;
  for (int i = size - 1; i >= 0; --i) {
    array_seq[i] = val;
    val = gscan_f(val, i);
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;

  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_gscanr:OK" << std::endl;
}    
		       

void test_shiftl()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int* array_skel = new int[size];
  {
    sketo::dist_list<int> as = sketo::list_skeletons::shiftl(11,
				 sketo::list_skeletons::shiftl(10,
				   sketo::list_skeletons::generate(size, sketo::functions::identity<int>())));
    as.gather(array_skel);
  }

  // sequential code
  int *array_seq = new int[size];
  for (int i = 0; i < size - 2; ++i) {
    array_seq[i] = i + 2;
  }
  array_seq[size - 2] = 10;
  array_seq[size - 1] = 11;
  
  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_shiftl:OK" << std::endl;
}


void test_shiftr()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int* array_skel = new int[size];
  {
    sketo::dist_list<int> as = sketo::list_skeletons::shiftr(11,
				 sketo::list_skeletons::shiftr(10,
				   sketo::list_skeletons::generate(size, sketo::functions::identity<int>())));
    as.gather(array_skel);
  }

  // sequential code
  int *array_seq = new int[size];
  for (int i = 2; i < size; ++i) {
    array_seq[i] = i - 2;
  }
  array_seq[0] = 11;
  array_seq[1] = 10;
  
  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_shiftr:OK" << std::endl;
}


void test_shiftl2()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int_vector* array_skel = new int_vector[size];
  {
    sketo::dist_list<int_vector> as = sketo::list_skeletons::shiftl(to_int_vector(11),
                                                                    sketo::list_skeletons::shiftl(to_int_vector(10),
				   sketo::list_skeletons::generate(size, to_int_vector)));
    as.gather(array_skel);
  }

  // sequential code
  int_vector *array_seq = new int_vector[size];
  for (int i = 0; i < size - 2; ++i) {
    array_seq[i] = to_int_vector(i + 2);
  }
  array_seq[size - 2] = to_int_vector(10);
  array_seq[size - 1] = to_int_vector(11);
  
  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_shiftl2:OK" << std::endl;
}


void test_shiftr2()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int_vector* array_skel = new int_vector[size];
  {
    sketo::dist_list<int_vector> as = sketo::list_skeletons::shiftr(to_int_vector(11),
                                                                    sketo::list_skeletons::shiftr(to_int_vector(10),
				   sketo::list_skeletons::generate(size, to_int_vector)));
    as.gather(array_skel);
  }

  // sequential code
  int_vector *array_seq = new int_vector[size];
  for (int i = 2; i < size; ++i) {
    array_seq[i] = to_int_vector(i - 2);
  }
  array_seq[0] = to_int_vector(11);
  array_seq[1] = to_int_vector(10);
  
  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_shiftr2:OK" << std::endl;
}

// 
// void test_accumulate()
// {
//   MPI_Barrier(MPI_COMM_WORLD);
//   // initial list
//   int *array = new int[size];
//   for (int i = 0; i < size; ++i) { array[i] = i + 1; }
    
//   // skeleton code
//   sketo::dist_list<int> *as = new sketo::dist_list<int>(size, array);

//   double_pair result_skel = sketo::list_skeletons::accumulate(accumulate_g, accumulate_p, accumulate_oplus, accumulate_q, std::multiplies<double>(), 10.0, as);

//   // sequential code
//   double *array_seq = new double[size];
//   array_seq[0] = 10.0;
//   for (int i = 1; i < size; ++i) { array_seq[i] = array_seq[i-1] * (-array[i-1]); }
//   double last = array_seq[size-1] * (-array[size-1]);
//   double_pair val(last, last);
//   for (int i = size - 1; i >= 0; --i) {
//     val.first -= 2 * array[i];
//     val.second += array_seq[i];
//   }
    
//   // compare
// //   std::cout << "LHS = " << result_skel << ", RHS = " << val << std::endl;
//   if (sketo::is_master()) { assert(result_skel == val); }

//   delete [] array_seq;
//   delete [] array;
//   delete as;

//   MPI_Barrier(MPI_COMM_WORLD);
//   sketo::cout << __FILE__ << ":test_accumulate:OK" << std::endl;
// }

#ifdef __GXX_EXPERIMENTAL_CXX0X__  // using c++0x
void test_lambda() {
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int result_skel = sketo::list_skeletons::reduce(
    [](int a, int b) { return a + b;},
    sketo::list_skeletons::map(
      [](int a) { return a * a;},
      sketo::list_skeletons::generate(size,
                                      [](int a) { return a;})));

  // sequential code
  int result_seq = 0;
  for (int i = 1; i < size; ++i) { result_seq += i*i; }

  // compare
  if (sketo::is_master()) assert(result_skel == result_seq);

  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_lambda:OK" << std::endl;
}

void test_auto_fusion() {
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code

  auto x1 = sketo::list_skeletons::generate(size, [](int a) { return a;});
  auto x2 = sketo::list_skeletons::map([](int a) { return a * a;}, x1);
  int result_skel = sketo::list_skeletons::reduce(
    [](int a, int b) { return a + b;}, x2);

  // sequential code
  int result_seq = 0;
  for (int i = 1; i < size; ++i) { result_seq += i*i; }

  // compare
  if (sketo::is_master()) assert(result_skel == result_seq);

  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_auto_fusion:OK" << std::endl;
}
#endif



int sketo::main(int, char**)
{
  test_get_set();
  test_get_set2();
  test_distribute_gather1();
  test_distribute_gather2();
  test_distribute_gather3();
  
  test_generate();
  test_map();
  test_map_fusion();
  test_map_with_index();
  test_zip();
  test_zipwith();
  test_reduce();
  test_scan();
  test_scanr();
  test_reduce2();
  test_scan2();
  test_scanr2();
  test_postscan();
  test_postscanr();
  test_gscanl();
  test_gscanr();
  test_shiftl();
  test_shiftr();
  test_shiftl2();
  test_shiftr2();
//   test_accumulate();
  
// to test the lambda expression, we should compile the test program by hand:
// /usr/local/mpich2-1.2.1p1/bin//mpicxx -L../src -lsketo -I../include -O3 test_list_skeletons.cpp -o test_list_skeletons.sketo -std=gnu++0x

#ifdef __GXX_EXPERIMENTAL_CXX0X__  // using c++0x
  test_lambda();
  test_auto_fusion();
#endif

  return 0;
}
